@@ -1,7 +1,7 @@
 /*
  * vim:ts=4:sw=4:expandtab
  *
- * © 2010 Michael Stapelberg
+ * © 2010-2011 Michael Stapelberg
  *
  * See LICENSE for licensing information
  *
@@ -16,8 +16,6 @@
 #include <xcb/dpms.h>
 #include <xcb/xcb_keysyms.h>
 #include <err.h>
-#include <cairo.h>
-#include <cairo/cairo-xcb.h>
 #include <assert.h>
 #include <security/pam_appl.h>
 /* FIXME: can we get rid of this header? */
@@ -25,15 +23,19 @@
 #include <getopt.h>
 #include <string.h>
 
+#ifndef NOLIBCAIRO
+#include <cairo.h>
+#include <cairo/cairo-xcb.h>
+#endif
+
 #include "keysym2ucs.h"
 #include "ucs2_to_utf8.h"
 #include "xcb.h"
 #include "cursors.h"
 
 static xcb_connection_t *conn;
+static xcb_cursor_t cursor;
 static xcb_key_symbols_t *symbols;
-static cairo_surface_t *img = NULL;
-static cairo_t *ctx = NULL;
 static xcb_screen_t *scr;
 static pam_handle_t *pam_handle;
 static int input_position = 0;
@@ -43,7 +45,12 @@ static bool modeswitch_active = false;
 static int modeswitchmask;
 static int numlockmask;
 static bool beep = false;
+
+#ifndef NOLIBCAIRO
+static cairo_surface_t *img = NULL;
+static cairo_t *ctx = NULL;
 static bool tile = false;
+#endif
 
 static void input_done() {
     if (input_position == 0)
@@ -55,6 +62,8 @@ static void input_done() {
         exit(0);
     }
 
+    fprintf(stderr, "Authentication failure\n");
+
     /* beep on authentication failure, if enabled */
     if (beep) {
         xcb_bell(conn, 100);
@@ -67,6 +76,7 @@ static void input_done() {
  *
  */
 static void handle_expose_event() {
+#ifndef NOLIBCAIRO
     if (!ctx)
         return;
 
@@ -82,7 +92,7 @@ static void handle_expose_event() {
         /* otherwise, just paint the image */
         cairo_paint(ctx);
     }
-
+#endif
     xcb_flush(conn);
 }
 
@@ -113,26 +123,22 @@ static void handle_key_release(xcb_key_release_event_t *event) {
 static void handle_key_press(xcb_key_press_event_t *event) {
     //printf("keypress %d, state raw = %d\n", event->detail, event->state);
 
-    /* fix state */
-    if (modeswitch_active)
-            event->state |= modeswitchmask;
-
-    /* Apparantly, after activating numlock once, the numlock modifier
-     * stays turned on (use xev(1) to verify). So, to resolve useful
-     * keysyms, we remove the numlock flag from the event state */
-    event->state &= ~numlockmask;
-
-    if ((input_position + 8) >= sizeof(password))
-        return;
-
-    xcb_keysym_t sym = xcb_key_press_lookup_keysym(symbols, event, event->state);
-    switch (sym) {
+    xcb_keysym_t sym0, sym1, sym;
+    if (modeswitch_active) {
+        sym0 = xcb_key_press_lookup_keysym(symbols, event, 4);
+        sym1 = xcb_key_press_lookup_keysym(symbols, event, 5);
+    } else {
+        sym0 = xcb_key_press_lookup_keysym(symbols, event, 0);
+        sym1 = xcb_key_press_lookup_keysym(symbols, event, 1);
+    }
+    switch (sym0) {
     case XK_Mode_switch:
         //printf("Mode switch enabled\n");
         modeswitch_active = true;
         return;
 
     case XK_Return:
+    case XK_KP_Enter:
         input_done();
     case XK_Escape:
         input_position = 0;
@@ -146,10 +152,24 @@ static void handle_key_press(xcb_key_press_event_t *event) {
         /* decrement input_position to point to the previous glyph */
         u8_dec(password, &input_position);
         password[input_position] = '\0';
-        printf("new input position = %d, new password = %s\n", input_position, password);
+        //printf("new input position = %d, new password = %s\n", input_position, password);
         return;
     }
 
+    if ((input_position + 8) >= sizeof(password))
+        return;
+
+    if ((event->state & numlockmask) && xcb_is_keypad_key(sym1)) {
+        /* this key was a keypad key */
+        if ((event->state & XCB_MOD_MASK_SHIFT))
+            sym = sym0;
+        else sym = sym1;
+    } else {
+        if ((event->state & XCB_MOD_MASK_SHIFT))
+            sym = sym1;
+        else sym = sym0;
+    }
+
 #if 0
     /* FIXME: handle all of these? */
     printf("is_keypad_key = %d\n", xcb_is_keypad_key(sym));
@@ -162,15 +182,15 @@ static void handle_key_press(xcb_key_press_event_t *event) {
 #endif
 
     if (xcb_is_modifier_key(sym) || xcb_is_cursor_key(sym))
-            return;
+        return;
 
     //printf("sym = %c (%d)\n", sym, sym);
 
     /* convert the keysym to UCS */
     uint16_t ucs = keysym2ucs(sym);
     if ((int16_t)ucs == -1) {
-            fprintf(stderr, "Keysym could not be converted to UCS, skipping\n");
-            return;
+        fprintf(stderr, "Keysym could not be converted to UCS, skipping\n");
+        return;
     }
 
     /* store the UCS in a string to convert it */
@@ -201,6 +221,25 @@ void handle_visibility_notify(xcb_visibility_notify_event_t *event) {
     }
 }
 
+/*
+ * Called when the keyboard mapping changes. We update our symbols and re-grab
+ * pointer/keyboard.
+ *
+ */
+void handle_mapping_notify(xcb_mapping_notify_event_t *event) {
+    printf("mapping notify\n");
+    xcb_refresh_keyboard_mapping(symbols, event);
+
+    xcb_ungrab_pointer(conn, XCB_CURRENT_TIME);
+    xcb_ungrab_keyboard(conn, XCB_CURRENT_TIME);
+    grab_pointer_and_keyboard(conn, scr, cursor);
+
+    modeswitchmask = get_mod_mask(conn, symbols, XK_Mode_switch);
+    numlockmask = get_mod_mask(conn, symbols, XK_Num_Lock);
+
+    xcb_flush(conn);
+}
+
 /*
  * Callback function for PAM. We only react on password request callbacks.
  *
@@ -238,14 +277,15 @@ int main(int argc, char *argv[]) {
     bool dpms = false;
     char color[7] = "ffffff";
     char *username;
+#ifndef NOLIBCAIRO
     char *image_path = NULL;
+#endif
     int ret;
     struct pam_conv conv = {conv_callback, NULL};
     int screen;
     xcb_visualtype_t *vistype;
     xcb_generic_event_t *event;
     xcb_window_t win;
-    xcb_cursor_t cursor;
     int curs_choice = CURS_NONE;
     char o;
     int optind = 0;
@@ -254,20 +294,26 @@ int main(int argc, char *argv[]) {
         {"nofork", no_argument, NULL, 'n'},
         {"beep", no_argument, NULL, 'b'},
         {"dpms", no_argument, NULL, 'd'},
-        {"image", required_argument, NULL, 'i'},
         {"color", required_argument, NULL, 'c'},
-        {"tiling", no_argument, NULL, 't'},
         {"pointer", required_argument, NULL , 'p'},
+#ifndef NOLIBCAIRO
+        {"image", required_argument, NULL, 'i'},
+        {"tiling", no_argument, NULL, 't'},
+#endif
         {NULL, no_argument, NULL, 0}
     };
 
     if ((username = getenv("USER")) == NULL)
         errx(1, "USER environment variable not set, please set it.\n");
 
-    while ((o = getopt_long(argc, argv, "vnbdi:c:tp:", longopts, &optind)) != -1) {
+    while ((o = getopt_long(argc, argv, "vnbdc:p:"
+#ifndef NOLIBCAIRO
+        "i:t"
+#endif
+        , longopts, &optind)) != -1) {
         switch (o) {
         case 'v':
-            errx(EXIT_SUCCESS, "version " VERSION " © 2010 Michael Stapelberg\n");
+            errx(EXIT_SUCCESS, "version " VERSION " © 2010-2011 Michael Stapelberg\n");
         case 'n':
             dont_fork = true;
             break;
@@ -277,9 +323,6 @@ int main(int argc, char *argv[]) {
         case 'd':
             dpms = true;
             break;
-        case 'i':
-            image_path = strdup(optarg);
-            break;
         case 'c': {
             char *arg = optarg;
 
@@ -292,9 +335,14 @@ int main(int argc, char *argv[]) {
 
             break;
         }
+#ifndef NOLIBCAIRO
+        case 'i':
+            image_path = strdup(optarg);
+            break;
         case 't':
             tile = true;
             break;
+#endif
         case 'p':
             if (!strcmp(optarg, "win")) {
                 curs_choice = CURS_WIN;
@@ -304,7 +352,13 @@ int main(int argc, char *argv[]) {
             }
             break;
         default:
-            errx(1, "i3lock: Unknown option. Syntax: i3lock [-v] [-n] [-b] [-d] [-i image.png] [-c color] [-t] [-p win|default]\n");
+            errx(1, "i3lock: Unknown option. Syntax: i3lock [-v] [-n] [-b] [-d] [-c color] [-p win|default]"
+#ifndef NOLIBCAIRO
+            " [-i image.png] [-t]"
+#else
+            " (compiled with NOLIBCAIRO)"
+#endif
+            "\n");
         }
     }
 
@@ -314,8 +368,9 @@ int main(int argc, char *argv[]) {
         errx(EXIT_FAILURE, "PAM: %s\n", pam_strerror(pam_handle, ret));
 
     /* Initialize connection to X11 */
-    if ((conn = xcb_connect(NULL, &screen)) == NULL)
-        err(EXIT_FAILURE, "xcb_connect()");
+    if ((conn = xcb_connect(NULL, &screen)) == NULL ||
+        xcb_connection_has_error(conn))
+        errx(EXIT_FAILURE, "Could not connect to X11, maybe you need to set DISPLAY?");
 
     if (!dont_fork) {
         /* In the parent process, we exit */
@@ -345,13 +400,13 @@ int main(int argc, char *argv[]) {
 
     grab_pointer_and_keyboard(conn, scr, cursor);
 
-    if (image_path)
-        img = cairo_image_surface_create_from_png(image_path);
-
     symbols = xcb_key_symbols_alloc(conn);
     modeswitchmask = get_mod_mask(conn, symbols, XK_Mode_switch);
     numlockmask = get_mod_mask(conn, symbols, XK_Num_Lock);
 
+#ifndef NOLIBCAIRO
+    if (image_path)
+        img = cairo_image_surface_create_from_png(image_path);
 
     if (img) {
         /* Initialize cairo */
@@ -364,6 +419,7 @@ int main(int argc, char *argv[]) {
 
         handle_expose_event();
     }
+#endif
 
     if (dpms)
         dpms_turn_off_screen(conn);
@@ -401,6 +457,11 @@ int main(int argc, char *argv[]) {
             continue;
         }
 
+        if (type == XCB_MAPPING_NOTIFY) {
+            handle_mapping_notify((xcb_mapping_notify_event_t*)event);
+            continue;
+        }
+
         printf("WARNING: unhandled event of type %d\n", type);
     }
 
